#!/usr/bin/python3 -u
import os
from aegon.libutils import readxyzs, writexyzs, rename, cutter_energy
from aegon.libcalc_lj    import opt_LJ_parallel
from growpal.libgrowpal  import growpal_parallel, display_info
from growpal.libdisc_usr import comparator_usr_batch
from growpal.libselect_clustering import selectbyclustering_comb_prom
#------------------------------------------------------------------------------------------------
nproc = 26
growatom='Mo'
ecut=50.0
#rule0=[150, 150, 150, 150, 150, 100, 100, 50]
rule=[150, 150, 150, 150, 150, 150, 100, 100, 100, 100, 100, 100]
#------------------------------------------------------------------------------------------
molecu0=False
if __name__ == "__main__":
    for iii in range(4, 150+1):
        base_name = 'LJ'+str(iii).zfill(3)
        if os.path.isfile(base_name+'.xyz'):
            print ("%s exists" %(base_name))
            anterior=base_name+'.xyz'
            continue
        if not molecu0:
            molecu0 = readxyzs(anterior)

        molecu0 = rename(molecu0, base_name, 5)
        molecu1 = growpal_parallel(molecu0, growatom, dtol=1.2, n_cores=nproc)
        #----------OPT----------------
        for batch_size in [100, 150, 100, 150, 100]:
            molecu0 = parallel_opt_LJ(molecu1, n_jobs=nproc)
            molecu1 = comparator_usr_batch(molecu0, batch_size=batch_size, tols=0.99, tole=0.1, mono=True)
        #--------------------------------------------
        molecu1=cutter_energy(molecu1, ecut)
        molecu0=selectbyclustering_comb_prom(molecu1, rule, "MBTR_dis", "MBTR_cos")
        display_info(molecu0[0:5], base_name)
        writexyzs(molecu0, base_name+'.xyz')



import numpy as np
from joblib import Parallel, delayed
from dscribe.descriptors import CoulombMatrix, ACSF, SOAP, MBTR
#------------------------------------------------------------------------------
def descriptors_MBTR_dis (lista):
    des_list = []
    geometria = {"function": "distance"}
    grid = {"min": 0, "max": 10, "sigma": 1E-2, "n": 200}
    pesaje = {"function": "inverse_square", "r_cut": 10, "threshold" :1e-3}
    for imol in lista:
         atoms = imol.get_chemical_symbols()
         especies = set(atoms)
         mbtr = MBTR(species = especies, geometry = geometria, grid = grid,
                     weighting = pesaje, periodic = False, normalization = "none",
                     normalize_gaussians = True, sparse = False, dtype = "float64")
         des = mbtr.create(imol)
         des_list.append(des)
    return des_list

def descriptors_MBTR_inv_dis (lista):
    des_list = []
    geometria = {"function": "inverse_distance"}
    grid = {"min": 0, "max": 1, "sigma": 0.1, "n" : 200}
    pesaje = {"function": "inverse_square", "r_cut": 10, "threshold" :1e-3}
    for  imol in lista:
        atoms = imol.get_chemical_symbols()
        especies = set(atoms)
        mbtr = MBTR(species = especies, geometry = geometria, grid = grid,
                    weighting = pesaje, periodic = False, normalization = "none",
                    normalize_gaussians = True, sparse = False, dtype = "float64")
        
        des = mbtr.create(imol)
        des_list.append(des)
    return des_list

def descriptors_MBTR_ang(lista):
    des_list = []
    geometria = {"function": "angle"}
    grid = {"min": 0, "max": 185, "sigma": 0.1, "n": 360}
    pesaje = {"function": "exp", "scale": 0.5, "threshold" :1e-3}
    for imol in lista:
        atoms = imol.get_chemical_symbols()
        especies = set(atoms)
        mbtr = MBTR(species = especies, geometry = geometria, grid = grid,
                    weighting = pesaje, periodic = False, normalization = "none",
                    normalize_gaussians = True, sparse = False, dtype = "float64")
        des = mbtr.create(imol)
        des_list.append(des)
    return des_list

def descriptors_MBTR_cos (lista):
    des_list = []
    geometria = {"function": "cosine"}
    grid = {"min": -1.05, "max": 1.05, "sigma": 1E-2, "n" : 200}
    pesaje = {"function": "exp", "scale": 0.5, "threshold" :1e-3}
    for imol in lista:
        atoms = imol.get_chemical_symbols()
        especies = set(atoms)
        mbtr = MBTR(species = especies, geometry = geometria, grid = grid,
                    weighting = pesaje, periodic = False, normalization = "none",
                    normalize_gaussians = True, sparse = False, dtype = "float64")
        des = mbtr.create(imol)
        des_list.append(des)
    return des_list


#------------------------------------------------------------------------------
def descriptors_ACSF (lista):
    des_list = []
    r_cut = 3.5
    g2params=[[1, 1], [1, 2], [1, 3]]
    g4params=[[1, 1, 1], [1, 2, 1], [1, 1, -1], [1, 2, -1]]
    for imol in lista:
         atoms = imol.get_chemical_symbols()
         especies = set(atoms)
         acsf = ACSF(species = especies, r_cut = r_cut, g2_params=g2params, g4_params=g4params)
         acsf_des = acsf.create(imol)
         des = acsf_des.flatten()
         des_list.append(des)
    return des_list

def descriptors_SOAP (lista):
    des_list = []
    r_cut = 6.0
    n_max = 8
    l_max = 6
    for imol in lista:
        atoms = imol.get_chemical_symbols()
        especies = set(atoms)
        soap = SOAP(species = especies, periodic=False, r_cut = r_cut, n_max=n_max, l_max=l_max)
        soap_des = soap.create(imol)
        des = soap_des.flatten()
        des_list.append(des)
    return des_list

def descriptors_CoulombMatrix (lista):
    des_list = []
    for imol in lista:
        filas = len(imol)
        cm = CoulombMatrix(n_atoms_max = filas)
        des = cm.create(imol)
        des_list.append(des)
    return des_list
#------------------------------------------------------------------------------
def split_list(lista, chunk_size):
    return [lista[i:i + chunk_size] for i in range(0, len(lista), chunk_size)]

def descriptors_sel(lista, des, chunk_size=500, n_jobs=-1):
    
    descriptor_functions = {"MBTR_dis": descriptors_MBTR_dis, "MBTR_inv_dis": descriptors_MBTR_inv_dis, 
                            "MBTR_ang": descriptors_MBTR_ang, "MBTR_cos": descriptors_MBTR_cos,
                            "ACSF": descriptors_ACSF, "SOAP": descriptors_SOAP,
                            "CoulombMatrix": descriptors_CoulombMatrix}
    
    des_fun = descriptor_functions[des]
    sublistas = split_list(lista, chunk_size)
    resultados = Parallel(n_jobs=n_jobs)(delayed(des_fun)(sublista) for sublista in sublistas)
    descriptores = [descriptor for sublista in resultados for descriptor in sublista]
    return descriptores

def descriptors_comb(lista, des_1, des_2, chunk_size=500, n_jobs=-1): 
    des_list1 = descriptors_sel(lista, des_1, chunk_size, n_jobs)
    des_list2 = descriptors_sel(lista, des_2, chunk_size, n_jobs)
    combined = [np.concatenate((d1, d2)) for d1, d2 in zip(des_list1, des_list2)]
    return combined



#!/usr/bin/python3 -u
import os.path
import numpy as np
from aegon.libutils import readxyzs
#------------------------------------------------------------------------------------------
def LennardJones(moleculein):
    epsilon=1.0
    sigma=3.0/np.power(2,1/6)
    energy=0.0
    natoms=len(moleculein)
    for iatom in range(natoms):
        ipos=moleculein[iatom].position
        for jatom in range(iatom+1,natoms):
            jpos = moleculein[jatom].position
            rij = np.linalg.norm(jpos - ipos)
            arg1=(sigma/rij)
            arg6=np.power(arg1,6)
            energy += arg6 * (arg6 - 1.0)
    energy=energy*4.0*epsilon
    return energy
#------------------------------------------------------------------------------------------
ljenergy={}
ljnet=readxyzs('Wales003to150.xyz')
for imol in ljnet:
    ljenergy[len(imol)]=imol.info['e']

ii=4
lista=[]
file='LJ'+str(ii).zfill(3)+'.xyz'
title = "resultados.txt"

if os.path.exists(title):
    os.remove(title)

with open(title, 'a') as f_out:
    while os.path.isfile(file):
        mol=readxyzs(file)
        nm=len(mol)
        #egulp=mol[0].e
        id=mol[0].info['i']
        nn=id[6:11]
        efili=LennardJones(mol[0])
        etrue=ljenergy[ii]
        deltae=(efili-etrue)
        pg='X'
        result_line = '%s (%4d) %11.6f %5.2f %-3s ISO%s\n' % (file[2:6], nm, efili, deltae, pg, nn)
        f_out.write(result_line)
        ii=ii+1
        file='LJ'+str(ii).zfill(3)+'.xyz'
exit()
#------------------------------------------------------------------------------------------



import os
import time
import numpy as np
from ase import Atom
from multiprocessing import Pool, cpu_count
from aegon.libpymatgen import inequivalent_finder
from aegon.libutils import rename, adjacency_matrix, readxyzs, writexyzs, sort_by_energy
#------------------------------------------------------------------------------------------
def neighbor_finder(atoms, dtol=1.2):
    """
    Build adjacency dictionary using list comprehension
    """
    dict_neig = {}
    adj_mtx = adjacency_matrix(atoms, dtol)
    for i, row in enumerate(adj_mtx):
        dict_neig[i] = [j for j, val in enumerate(row) if val == 1]
    return dict_neig

#------------------------------------------------------------------------------------------
def triangle_finder(neighbors_dict, atom):
    """
    Optimized version using sets for O(1) lookups
    """
    triangles = []
    neiga = neighbors_dict[atom]
    set_neiga = set(neiga)
    
    for n in neiga:
        neigb = neighbors_dict[n]
        for n2 in neigb:
            if n2 in set_neiga:
                auxlist = [atom, n, n2]
                auxlist.sort()
                if auxlist not in triangles:
                    triangles.append(auxlist)
    return triangles

#------------------------------------------------------------------------------------------
def focused_expansion(molin, vector):
    """
    Vectorized version
    """
    molout = molin.copy()
    positions = molout.positions
    vectors = positions - vector
    distances = np.linalg.norm(vectors, axis=1, keepdims=True)
    
    distances = np.where(distances == 0, 1e-10, distances)
    
    rnorm = vectors / distances
    edel = np.exp(-distances) * 1.2
    new_positions = positions + rnorm * edel
    
    molout.positions = new_positions
    return molout

#------------------------------------------------------------------------------------------
def add_all_interstitial_atoms(original_molecule, neighbors_dict, specie):
    """
    Optimized version using tuples and sets for visited_t
    """
    visited_t = set()
    xxx_mol = original_molecule.copy()
    atom_list = range(len(original_molecule))
    
    all_positions = original_molecule.positions
    
    for a in atom_list:
        triangles = triangle_finder(neighbors_dict, a)
        for t in triangles:
            t_tuple = tuple(t)
            if t_tuple in visited_t:
                continue
            
            v1 = all_positions[t[0]]
            v2 = all_positions[t[1]] 
            v3 = all_positions[t[2]]
            mid_vect = (v1 + v2 + v3) / 3
            
            add_atom = Atom(symbol=specie, position=mid_vect)
            xxx_mol.append(add_atom)
            visited_t.add(t_tuple)
            
    return xxx_mol

#------------------------------------------------------------------------------------------
def add_ineq_interstitial_atoms_with_expansion(original_molecule, atom_list):
    """
    Process eligible atoms with expansion
    """
    molist_out = []
    for add_atom in atom_list:
        mid_vect = np.array(add_atom.position)
        add_mol = original_molecule.copy()
        exp_mol = focused_expansion(add_mol, mid_vect)
        exp_mol.append(add_atom)
        molist_out.append(exp_mol)
    return molist_out

#------------------------------------------------------------------------------------------
def process_single_molecule(args):
    """
    Process one molecule - designed for parallel execution
    """
    imol, specie, dtol = args
    org_nnn = len(imol)
    org_mol = imol.copy()
    org_mol.info['e'] = 0.0
    org_mol.translate(-org_mol.get_center_of_mass())
    
    all_neighbors = neighbor_finder(org_mol, dtol)
    xxx_mol = add_all_interstitial_atoms(org_mol, all_neighbors, specie)
    inequivalent = inequivalent_finder(xxx_mol)
    elegible_atoms = [xxx_mol[i] for i in inequivalent if i >= org_nnn]
    mod_mol = add_ineq_interstitial_atoms_with_expansion(org_mol, elegible_atoms)
    mod_mol = rename(mod_mol, imol.info['i'], 4)
    
    return mod_mol

#------------------------------------------------------------------------------------------
def growpal_parallel(poscarlist, specie, dtol=1.2, n_cores=None):
    """
    Parallel version using multiprocessing Pool
    """
    start = time.time()
    
    if n_cores is None:
        n_cores = max(1, cpu_count() - 1)
    
    args_list = [(imol, specie, dtol) for imol in poscarlist]
    
    with Pool(processes=n_cores) as pool:
        results = pool.map(process_single_molecule, args_list)
    
    molist_out = []
    for result in results:
        molist_out.extend(result)
    
    end = time.time()
    print('GrowPal generation at %5.2f s' % (end - start))
    
    return molist_out

#------------------------------------------------------------------------------------------
def display_info(moleculein, stage_string):
    print("-------------------------------------------------------------------")
    print("------------------------- SUMMARY %s -------------------------" %(stage_string))
    print("Number File--------Name   Energy (ev)   Delta----E T")
    molzz = sort_by_energy(moleculein, 1)
    emin = molzz[0].info['e']
    for ii, imol in enumerate(molzz):
        ei = imol.info['e']
        id = imol.info['i']
        nt = imol.info['c']
        deltae  =  ei - emin
        kk=str(ii+1).zfill(6)
        print("%s %s %13.8f %12.8f %d" %(kk, id, ei, deltae, nt))
#------------------------------------------------------------------------------------------
#if __name__ == "__main__":
#    mol1 = readxyzs('../LJ012.xyz')
#    mol2 = growpal_parallel(mol1, 'Nb')
#    writexyzs(mol2, 'out.xyz')



import os
import time
import numpy as np
from numba import jit
import multiprocessing as mp
from ase.data import atomic_masses
from aegon.libutils import readxyzs, writexyzs, centroid, sort_by_energy, align_two
#------------------------------------------------------------------------------------------
# Global tolerances
tolsij = 0.95
tolene = 0.10
#------------------------------------------------------------------------------------------
# Optimized USR helper functions
#------------------------------------------------------------------------------------------
@jit(nopython=True)
def lastwo_central_moment_numba(arr):
    """
    Compute the last two central moments (variance and third central moment)
    for a 1D numpy array using Numba for speed.
    Returns: array([second_moment, third_moment])
    """
    xavg = np.mean(arr)
    res = arr - xavg
    ssq2 = np.mean(res**2)
    ssq3 = np.mean(res**3)
    return np.array([ssq2, ssq3])
#------------------------------------------------------------------------------------------
def lastwo_central_moment(listxxx):
    """
    Wrapper that converts an arbitrary sequence to float64 numpy array
    and calls the Numba implementation.
    """
    arr = np.array(listxxx, dtype=np.float64)
    return lastwo_central_moment_numba(arr)
#------------------------------------------------------------------------------------------
@jit(nopython=True)
def find_extreme_points_numba(positions, centroid):
    """
    Find four characteristic points used by USR:
      - centroid (provided)
      - atom closest to centroid
      - atom farthest from centroid
      - atom farthest from the farthest atom (to obtain an extended diameter)
    Implemented in Numba for speed. Returns a tuple:
      (centroid, closest_pos, farthest_pos, farthest_from_farthest_pos)
    """
    n_atoms = len(positions)
    distances_to_centroid = np.zeros(n_atoms)
    for i in range(n_atoms):
        distances_to_centroid[i] = np.linalg.norm(positions[i] - centroid)
    closest_idx = np.argmin(distances_to_centroid)
    farthest_idx = np.argmax(distances_to_centroid)

    distances_from_farthest = np.zeros(n_atoms)
    for i in range(n_atoms):
        distances_from_farthest[i] = np.linalg.norm(positions[i] - positions[farthest_idx])
    farthest_from_farthest_idx = np.argmax(distances_from_farthest)

    return (centroid, positions[closest_idx], positions[farthest_idx], positions[farthest_from_farthest_idx])
#------------------------------------------------------------------------------------------
def four_points(atoms):
    """
    Build the positions array and compute the four characteristic points
    using the centroid helper and the Numba extreme point finder.
    """
    positions = np.array([atom.position for atom in atoms])
    ctd = centroid(atoms)
    return find_extreme_points_numba(positions, ctd)
#------------------------------------------------------------------------------------------
@jit(nopython=True)
def calculate_usr_descriptors_numba(positions, points, masses=None, mass_avg=None):
    """
    Calculate USR descriptors for a set of positions given the four points.
    If masses is provided, compute mass-weighted moments as additional features.
    Returns:
      - 8 features for monoatomic (4 x [variance, third_moment])
      - 16 features if masses provided (adds 4 mass-weighted pairs)
    """
    ctd, cst, fct, ftf = points
    n_atoms = len(positions)
    lctd = np.zeros(n_atoms)
    lcst = np.zeros(n_atoms)
    lfct = np.zeros(n_atoms)
    lftf = np.zeros(n_atoms)

    for i in range(n_atoms):
        pos = positions[i]
        lctd[i] = np.linalg.norm(pos - ctd)
        lcst[i] = np.linalg.norm(pos - cst)
        lfct[i] = np.linalg.norm(pos - fct)
        lftf[i] = np.linalg.norm(pos - ftf)

    a1 = lastwo_central_moment_numba(lctd)
    a2 = lastwo_central_moment_numba(lcst)
    a3 = lastwo_central_moment_numba(lfct)
    a4 = lastwo_central_moment_numba(lftf)

    # If no masses are provided, return the unweighted descriptors (8 values)
    if masses is None:
        return np.concatenate((a1, a2, a3, a4))

    # Otherwise compute mass-weighted distances and their moments
    lctdm = np.zeros(n_atoms)
    lcstm = np.zeros(n_atoms)
    lfctm = np.zeros(n_atoms)
    lftfm = np.zeros(n_atoms)
    for i in range(n_atoms):
        mi = masses[i]
        lctdm[i] = lctd[i] * mi / mass_avg
        lcstm[i] = lcst[i] * mi / mass_avg
        lfctm[i] = lfct[i] * mi / mass_avg
        lftfm[i] = lftf[i] * mi / mass_avg

    b1 = lastwo_central_moment_numba(lctdm)
    b2 = lastwo_central_moment_numba(lcstm)
    b3 = lastwo_central_moment_numba(lfctm)
    b4 = lastwo_central_moment_numba(lftfm)

    return np.concatenate((a1, a2, a3, a4, b1, b2, b3, b4))
#------------------------------------------------------------------------------------------
def USRMonoatom(moleculein):
    """
    Compute the 8-feature USR descriptor (geometry-only) for a molecule.
    """
    positions = np.array([atom.position for atom in moleculein])
    points = four_points(moleculein)
    return calculate_usr_descriptors_numba(positions, points)
#------------------------------------------------------------------------------------------
def USRMultiatom(moleculein):
    """
    Compute the 16-feature USR descriptor including mass-weighted moments.
    """
    positions = np.array([atom.position for atom in moleculein])
    masses = np.array([atomic_masses[atom.number] for atom in moleculein])
    mass_avg = np.mean(masses)
    points = four_points(moleculein)
    return calculate_usr_descriptors_numba(positions, points, masses, mass_avg)
#------------------------------------------------------------------------------------------
@jit(nopython=True)
def similarity_numba(vi, vj, n_features):
    """
    Fast similarity metric implemented in Numba:
    similarity = 1 / (1 + (Manhattan_distance / n_features))
    Returns a float in (0,1].
    """
    manhattan_distance = np.sum(np.abs(vi - vj))
    return 1.0 / (1.0 + manhattan_distance / float(n_features))
#------------------------------------------------------------------------------------------
def compute_similarity_matrix_serial(descriptors, n_features):
    """
    Compute a symmetric similarity matrix (serial) for a list of descriptors.
    """
    n_mols = len(descriptors)
    similarity_matrix = np.zeros((n_mols, n_mols))
    for i in range(n_mols):
        vi = descriptors[i]
        for j in range(i, n_mols):
            vj = descriptors[j]
            sij = similarity_numba(vi, vj, n_features)
            similarity_matrix[i, j] = sij
            similarity_matrix[j, i] = sij
    return similarity_matrix
#------------------------------------------------------------------------------------------
def find_similar_elements(listmol, similarity_matrix, tols=tolsij, tole=tolene):
    """
    Identify and remove molecules similar by geometry (similarity >= tols)
    and close in energy (|DeltaE| <= tole). Returns list of dissimilar (kept) molecules.
    """
    similar_indices = []
    num_elements = similarity_matrix.shape[0]

    for i in range(num_elements):
        for j in range(i + 1, num_elements):
            edf = np.abs(listmol[i].info['e'] - listmol[j].info['e'])
            if (similarity_matrix[i, j] >= tols) and (edf <= tole):
                similar_indices.append(j)

    similar_indices = list(set(similar_indices))
    dissimilars_atoms = [listmol[i] for i in range(num_elements) if i not in similar_indices]
    return dissimilars_atoms
#------------------------------------------------------------------------------------------
def disc_USR_sublist(args):
    """
    Process a sublist of molecules with USR descriptors — worker for multiprocessing.
    Args tuple: (sublist, tols, tole, mono)
    Returns: filtered list for the sublist
    """
    sublist, tols, tole, mono = args
    num_molecules = len(sublist)
    if num_molecules == 0:
        return []

    n_features = 8 if mono else 16

    # Compute descriptors for the sublist
    if mono:
        descriptors = [USRMonoatom(mol) for mol in sublist]
    else:
        descriptors = [USRMultiatom(mol) for mol in sublist]
    # Compute similarity matrix and filter
    similarity_matrix = compute_similarity_matrix_serial(descriptors, n_features)
    return find_similar_elements(sublist, similarity_matrix, tols, tole)
#------------------------------------------------------------------------------------------
def split_list(lst, n):
    """
    Split a list into n approximately equal parts.
    """
    k, m = divmod(len(lst), n)
    return [lst[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n)]
#------------------------------------------------------------------------------------------
def _second_pass_filtering(molecules, tols, tole, mono):
    """
    Efficient second-pass filtering across many molecules.
    Only compares molecules close in energy to reduce comparisons.
    """
    if len(molecules) <= 1:
        return molecules

    molecules_sorted = sorted(molecules, key=lambda x: x.info['e'])
    n_mols = len(molecules_sorted)

    # Precompute descriptors once
    n_features = 8 if mono else 16
    if mono:
        descriptors = [USRMonoatom(mol) for mol in molecules_sorted]
    else:
        descriptors = [USRMultiatom(mol) for mol in molecules_sorted]

    keep_indices = set(range(n_mols))

    # For each molecule, compare only to neighbors in energy within a window
    for i in range(n_mols):
        if i not in keep_indices:
            continue

        vi = descriptors[i]
        ei = molecules_sorted[i].info['e']

        # Limit search range to save time (adjustable heuristic)
        search_range = min(50, n_mols // 10)
        start_j = max(0, i - search_range)
        end_j = min(n_mols, i + search_range + 1)

        for j in range(start_j, end_j):
            if j <= i or j not in keep_indices:
                continue

            edf = np.abs(ei - molecules_sorted[j].info['e'])
            if edf > tole:
                continue

            sij = similarity_numba(vi, descriptors[j], n_features)
            if sij >= tols:
                keep_indices.remove(j)

    return [molecules_sorted[i] for i in sorted(keep_indices)]
#------------------------------------------------------------------------------------------
def comparator_usr_serial(molecules, tols=tolsij, tole=tolene, mono=False):
    """
    Serial USR comparator — computes descriptors and filters duplicates entirely in memory.
    Returns filtered list of molecules.
    """
    ni = len(molecules)
    if ni == 0:
        return []

    n_features = 8 if mono else 16

    # Compute descriptors
    if mono:
        descriptors = [USRMonoatom(mol) for mol in molecules]
    else:
        descriptors = [USRMultiatom(mol) for mol in molecules]

    # Build similarity matrix and filter
    similarity_matrix = compute_similarity_matrix_serial(descriptors, n_features)
    filtered_molecules = find_similar_elements(molecules, similarity_matrix, tols, tole)

    return filtered_molecules
#------------------------------------------------------------------------------------------
def comparator_usr_batch(molecules, batch_size=100, tols=tolsij, tole=tolene, mono=False):
    """
    Batch-processing USR comparator (no multiprocessing).
    Useful when multiprocessing is unreliable or not desired.
    Processes molecules in batches, then applies a second-pass filter across batches.
    """
    start = time.time()
    ni = len(molecules)

    if ni == 0:
        return []

    molecules = sort_by_energy(molecules, 1)

    # If the list is small, run the serial comparator directly
    if ni <= batch_size:
        return comparator_usr_serial(molecules, tols, tole, mono)

    # Process in batches
    all_filtered_mols = []
    n_batches = (ni + batch_size - 1) // batch_size

    for i in range(n_batches):
        start_idx = i * batch_size
        end_idx = min((i + 1) * batch_size, ni)
        batch = molecules[start_idx:end_idx]

        filtered_batch = comparator_usr_serial(batch, tols, tole, mono)
        all_filtered_mols.extend(filtered_batch)

    # Second pass filtering between batches to remove cross-batch duplicates
    if len(all_filtered_mols) > 1:
        all_filtered_mols = _second_pass_filtering(all_filtered_mols, tols, tole, mono)

    # Sort final list by energy
    all_filtered_mols = sort_by_energy(all_filtered_mols, 1)

    nf = len(all_filtered_mols)
    end = time.time()
    print('USR comparison (batch) at %5.2f s [%d -> %d]' % (end - start, ni, nf))

    return all_filtered_mols

#------------------------------------------------------------------------------------------
# Additional functions for comparison against references
#------------------------------------------------------------------------------------------

def make_similarity_matrix_compare(moleculein, moleculeref, mono=False):
    """
    Build a cross-similarity matrix between two molecule lists.
    Returns a (len(moleculein) x len(moleculeref)) numpy array of similarities.
    """
    n_features = 8 if mono else 16

    # Compute descriptors for both sets
    if mono:
        descriptors1 = [USRMonoatom(mol) for mol in moleculein]
        descriptors2 = [USRMonoatom(mol) for mol in moleculeref]
    else:
        descriptors1 = [USRMultiatom(mol) for mol in moleculein]
        descriptors2 = [USRMultiatom(mol) for mol in moleculeref]

    total_molecules1, total_molecules2 = len(moleculein), len(moleculeref)
    similarity_matrix = np.zeros((total_molecules1, total_molecules2), dtype=float)

    # Fill the matrix
    for i in range(total_molecules1):
        vi = descriptors1[i]
        for j in range(total_molecules2):
            vj = descriptors2[j]
            similarity_matrix[i, j] = similarity_numba(vi, vj, n_features)

    return similarity_matrix
#------------------------------------------------------------------------------------------
def molin_sim_molref(moleculein, moleculeref, tols=tolsij, tole=tolene, mono=False):
    """
    Compare molecules in 'moleculein' against reference set 'moleculeref'.
    Removes molecules from 'moleculein' that are similar to any reference
    (similarity >= tols and |DeltaE| <= tole).
    Returns the filtered moleculein list.
    """
    start = time.time()
    ni = len(moleculein)

    matrixs = make_similarity_matrix_compare(moleculein, moleculeref, mono)

    similares = []
    for i, imol in enumerate(moleculein):
        for j, jmol in enumerate(moleculeref):
            if (matrixs[i, j] >= tols) and (np.abs(imol.info['e'] - jmol.info['e']) <= tole):
                similares.append(i)
                break

    moleculeout = [imol for i, imol in enumerate(moleculein) if i not in similares]
    nf = len(moleculeout)
    end = time.time()
    print('USR comparison (vs reference) at %5.2f s [%d -> %d]' % (end - start, ni, nf))
    return moleculeout
#------------------------------------------------------------------------------------------

